(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('worker-timers'), require('midi-file-slicer'), require('@babel/runtime/helpers/defineProperty'), require('json-midi-message-encoder'), require('@babel/runtime/helpers/classCallCheck'), require('@babel/runtime/helpers/createClass')) :
    typeof define === 'function' && define.amd ? define(['exports', 'worker-timers', 'midi-file-slicer', '@babel/runtime/helpers/defineProperty', 'json-midi-message-encoder', '@babel/runtime/helpers/classCallCheck', '@babel/runtime/helpers/createClass'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.midiPlayer = {}, global.workerTimers, global.midiFileSlicer, global._defineProperty, global.jsonMidiMessageEncoder, global._classCallCheck, global._createClass));
})(this, (function (exports, workerTimers, midiFileSlicer, _defineProperty, jsonMidiMessageEncoder, _classCallCheck, _createClass) { 'use strict';

    var createMidiFileSlicer = function createMidiFileSlicer(json) {
      return new midiFileSlicer.MidiFileSlicer({
        json: json
      });
    };

    var encodeMidiMessage = function encodeMidiMessage(event) {
      return new Uint8Array(jsonMidiMessageEncoder.encode(event));
    };

    var ALL_SOUND_OFF_EVENT_DATA = Array.from({
      length: 16
    }, function (_, index) {
      return new Uint8Array([176 + index, 120, 0]);
    });
    var MidiPlayer = /*#__PURE__*/function () {
      function MidiPlayer(_ref) {
        var encodeMidiMessage = _ref.encodeMidiMessage,
          filterMidiMessage = _ref.filterMidiMessage,
          json = _ref.json,
          midiFileSlicer = _ref.midiFileSlicer,
          midiOutput = _ref.midiOutput,
          startScheduler = _ref.startScheduler;
        _classCallCheck(this, MidiPlayer);
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._startScheduler = startScheduler;
        this._state = null;
      }
      return _createClass(MidiPlayer, [{
        key: "state",
        get: function get() {
          return this._state === null ? 'stopped' : this._state.stopScheduler === null ? 'paused' : 'playing';
        }
      }, {
        key: "pause",
        value: function pause() {
          if (this._state === null || this._state.stopScheduler === null) {
            throw new Error('The player is not playing.');
          }
          this._clear();
          var _this$_state = this._state,
            resolve = _this$_state.resolve,
            stopScheduler = _this$_state.stopScheduler;
          this._state.offset = stopScheduler();
          this._state.stopScheduler = null;
          resolve();
        }
      }, {
        key: "play",
        value: function play() {
          if (this._state !== null) {
            throw new Error('The player is not stopped.');
          }
          return this._schedule(0, 0);
        }
      }, {
        key: "resume",
        value: function resume() {
          if (this._state === null || this._state.stopScheduler !== null) {
            throw new Error('The player is not paused.');
          }
          var _this$_state2 = this._state,
            endedTracks = _this$_state2.endedTracks,
            offset = _this$_state2.offset;
          this._state = null;
          return this._schedule(endedTracks, offset);
        }
      }, {
        key: "stop",
        value: function stop() {
          if (this._state === null) {
            throw new Error('The player is already stopped.');
          }
          if (this._state.stopScheduler !== null) {
            this._clear();
            this._stop(this._state);
          }
        }
      }, {
        key: "_clear",
        value: function _clear() {
          var _this = this;
          var _a, _b;
          // Bug #1: Chrome does not yet implement the clear() method.
          (_b = (_a = this._midiOutput).clear) === null || _b === void 0 ? void 0 : _b.call(_a);
          ALL_SOUND_OFF_EVENT_DATA.forEach(function (data) {
            return _this._midiOutput.send(data);
          });
        }
      }, {
        key: "_schedule",
        value: function _schedule(endedTracks, offset) {
          var _this2 = this;
          return new Promise(function (resolve) {
            var stopScheduler = _this2._startScheduler(function (_ref2) {
              var end = _ref2.end,
                start = _ref2.start;
              if (_this2._state === null) {
                _this2._state = {
                  endedTracks: endedTracks,
                  offset: start - offset,
                  resolve: resolve,
                  stopScheduler: null
                };
              }
              var events = _this2._midiFileSlicer.slice(start - _this2._state.offset, end - _this2._state.offset);
              events.filter(function (_ref3) {
                var event = _ref3.event;
                return _this2._filterMidiMessage(event);
              }).forEach(function (_ref4) {
                var event = _ref4.event,
                  time = _ref4.time;
                return _this2._midiOutput.send(_this2._encodeMidiMessage(event), start + time);
              });
              _this2._state.endedTracks += events.filter(function (_ref5) {
                var event = _ref5.event;
                return MidiPlayer._isEndOfTrack(event);
              }).length;
              if (_this2._state.endedTracks === _this2._json.tracks.length) {
                _this2._stop(_this2._state);
              }
            });
            if (_this2._state === null) {
              stopScheduler();
            } else {
              _this2._state.stopScheduler = stopScheduler;
            }
          });
        }
      }, {
        key: "_stop",
        value: function _stop(state) {
          var resolve = state.resolve,
            stopScheduler = state.stopScheduler;
          stopScheduler === null || stopScheduler === void 0 ? void 0 : stopScheduler();
          this._state = null;
          resolve();
        }
      }], [{
        key: "_isEndOfTrack",
        value: function _isEndOfTrack(event) {
          return 'endOfTrack' in event;
        }
      }]);
    }();

    function ownKeys(e, r) { var t = Object.keys(e); if (Object.getOwnPropertySymbols) { var o = Object.getOwnPropertySymbols(e); r && (o = o.filter(function (r) { return Object.getOwnPropertyDescriptor(e, r).enumerable; })), t.push.apply(t, o); } return t; }
    function _objectSpread(e) { for (var r = 1; r < arguments.length; r++) { var t = null != arguments[r] ? arguments[r] : {}; r % 2 ? ownKeys(Object(t), !0).forEach(function (r) { _defineProperty(e, r, t[r]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(e, Object.getOwnPropertyDescriptors(t)) : ownKeys(Object(t)).forEach(function (r) { Object.defineProperty(e, r, Object.getOwnPropertyDescriptor(t, r)); }); } return e; }
    var createMidiPlayerFactory = function createMidiPlayerFactory(createMidiFileSlicer, startScheduler) {
      return function (options) {
        var midiFileSlicer = createMidiFileSlicer(options.json);
        return new MidiPlayer(_objectSpread(_objectSpread({
          filterMidiMessage: function filterMidiMessage(event) {
            return 'controlChange' in event || 'noteOff' in event || 'noteOn' in event || 'programChange' in event;
          }
        }, options), {}, {
          encodeMidiMessage: encodeMidiMessage,
          midiFileSlicer: midiFileSlicer,
          startScheduler: startScheduler
        }));
      };
    };

    var INTERVAL = 500;
    var createStartScheduler = function createStartScheduler(clearInterval, performance, setInterval) {
      return function (next) {
        var start = performance.now();
        var nextTick = start + INTERVAL;
        var end = nextTick + INTERVAL;
        var intervalId = setInterval(function () {
          if (performance.now() >= nextTick) {
            nextTick = end;
            end += INTERVAL;
            next({
              end: end,
              start: nextTick
            });
          }
        }, INTERVAL / 10);
        next({
          end: end,
          start: start
        });
        return function () {
          clearInterval(intervalId);
          return performance.now() - start;
        };
      };
    };

    var createMidiPlayer = createMidiPlayerFactory(createMidiFileSlicer, createStartScheduler(workerTimers.clearInterval, performance, workerTimers.setInterval));
    var create = function create(options) {
      return createMidiPlayer(options);
    };

    exports.create = create;

}));
